@@ -38,7 +38,9 @@
 #include "trap.h"
 
 #include "shell.h"
+#include "execute_cmd.h"
 #include "flags.h"
+#include "parser.h"
 #include "input.h"	/* for save_token_state, restore_token_state */
 #include "jobs.h"
 #include "signames.h"
@@ -85,19 +87,10 @@ static void reset_or_restore_signal_handlers (sh_resetsig_func_t *);
 static void trap_if_untrapped (int, char *);
 
 /* Variables used here but defined in other files. */
-extern int last_command_exit_value;
-extern int line_number;
-
-extern int sigalrm_seen;
 extern procenv_t alrmbuf;
 
 extern volatile int from_return_trap;
 
-extern char *this_command_name;
-extern sh_builtin_func_t *this_shell_builtin;
-extern procenv_t wait_intr_buf;
-extern int return_catch_flag, return_catch_value;
-extern int subshell_level;
 extern WORD_LIST *subst_assign_varlist;
 
 /* The list of things to do originally, before we started trapping. */
@@ -285,6 +278,7 @@ run_pending_traps ()
   register int sig;
   int old_exit_value, x;
   WORD_LIST *save_subst_varlist;
+  HASH_TABLE *save_tempenv;
   sh_parser_state_t pstate;
 #if defined (ARRAY_VARS)
   ARRAY *ps;
@@ -396,6 +390,8 @@ run_pending_traps ()
 	      save_parser_state (&pstate);
 	      save_subst_varlist = subst_assign_varlist;
 	      subst_assign_varlist = 0;
+	      save_tempenv = temporary_env;
+	      temporary_env = 0;	/* traps should not run with temporary env */
 
 #if defined (JOB_CONTROL)
 	      save_pipeline (1);	/* XXX only provides one save level */
@@ -409,6 +405,7 @@ run_pending_traps ()
 
 	      subst_assign_varlist = save_subst_varlist;
 	      restore_parser_state (&pstate);
+	      temporary_env = save_tempenv;
 	    }
 
 	  pending_traps[sig] = 0;	/* XXX - move before evalstring? */
@@ -457,7 +454,7 @@ trap_handler (sig)
       if (this_shell_builtin && (this_shell_builtin == wait_builtin))
 	{
 	  wait_signal_received = sig;
-	  if (interrupt_immediately)
+	  if (interrupt_immediately && wait_intr_flag)
 	    sh_longjmp (wait_intr_buf, 1);
 	}
 
@@ -489,6 +486,8 @@ first_pending_trap ()
   return -1;
 }
 
+/* Return > 0 if any of the "real" signals (not fake signals like EXIT) are
+   trapped. */
 int
 any_signals_trapped ()
 {
@@ -935,11 +934,12 @@ _run_trap_internal (sig, tag)
      char *tag;
 {
   char *trap_command, *old_trap;
-  int trap_exit_value, *token_state;
-  volatile int save_return_catch_flag, function_code, top_level_code, old_int;
+  int trap_exit_value;
+  volatile int save_return_catch_flag, function_code, old_int;
   int flags;
   procenv_t save_return_catch;
   WORD_LIST *save_subst_varlist;
+  HASH_TABLE *save_tempenv;
   sh_parser_state_t pstate;
 #if defined (ARRAY_VARS)
   ARRAY *ps;
@@ -976,6 +976,8 @@ _run_trap_internal (sig, tag)
       save_parser_state (&pstate);
       save_subst_varlist = subst_assign_varlist;
       subst_assign_varlist = 0;
+      save_tempenv = temporary_env;
+      temporary_env = 0;	/* traps should not run with temporary env */
 
 #if defined (JOB_CONTROL)
       if (sig != DEBUG_TRAP)	/* run_debug_trap does this */
@@ -994,9 +996,12 @@ _run_trap_internal (sig, tag)
       if (sig != DEBUG_TRAP && sig != RETURN_TRAP && sig != ERROR_TRAP)
 	flags |= SEVAL_RESETLINE;
       if (function_code == 0)
-	parse_and_execute (trap_command, tag, flags);
-
-      trap_exit_value = last_command_exit_value;
+        {
+	  parse_and_execute (trap_command, tag, flags);
+	  trap_exit_value = last_command_exit_value;
+        }
+      else
+        trap_exit_value = return_catch_value;
 
 #if defined (JOB_CONTROL)
       if (sig != DEBUG_TRAP)	/* run_debug_trap does this */
@@ -1010,6 +1015,8 @@ _run_trap_internal (sig, tag)
       restore_pipestatus_array (ps);
 #endif
 
+      temporary_env = save_tempenv;
+
       sigmodes[sig] &= ~SIG_INPROGRESS;
       running_trap = 0;
       interrupt_state = old_int;
@@ -1047,7 +1054,7 @@ _run_trap_internal (sig, tag)
 int
 run_debug_trap ()
 {
-  int trap_exit_value;
+  int trap_exit_value, old_verbose;
   pid_t save_pgrp;
   int save_pipe[2];
 
@@ -1065,8 +1072,15 @@ run_debug_trap ()
       stop_making_children ();
 #endif
 
+      old_verbose = echo_input_at_read;
+#if 0	/* not yet */
+      echo_input_at_read = 0;
+#endif
+
       trap_exit_value = _run_trap_internal (DEBUG_TRAP, "debug trap");
 
+      echo_input_at_read = old_verbose;
+
 #if defined (JOB_CONTROL)
       pipeline_pgrp = save_pgrp;
       restore_pipeline (1);
@@ -1074,8 +1088,9 @@ run_debug_trap ()
       close_pgrp_pipe ();
       restore_pgrp_pipe (save_pipe);
 #  endif
-      if (pipeline_pgrp > 0)
+      if (pipeline_pgrp > 0 && ((subshell_environment & (SUBSHELL_ASYNC|SUBSHELL_PIPE)) == 0))
 	give_terminal_to (pipeline_pgrp, 1);
+
       notify_and_cleanup ();
 #endif
       
@@ -1140,12 +1155,20 @@ free_trap_strings ()
 {
   register int i;
 
-  for (i = 0; i < BASH_NSIG; i++)
+  for (i = 0; i < NSIG; i++)
     {
       if (trap_list[i] != (char *)IGNORE_SIG)
 	free_trap_string (i);
     }
-  trap_list[DEBUG_TRAP] = trap_list[EXIT_TRAP] = trap_list[ERROR_TRAP] = trap_list[RETURN_TRAP] = (char *)NULL;
+  for (i = NSIG; i < BASH_NSIG; i++)
+    {
+      /* Don't free the trap string if the subshell inherited the trap */
+      if ((sigmodes[i] & SIG_TRAPPED) == 0)
+	{
+	  free_trap_string (i);
+	  trap_list[i] = (char *)NULL;
+	}
+    }
 }
 
 /* Free a trap command string associated with SIG without changing signal
